use crate::vm::dvm_value::DvmValue;
use crate::vm::mem::get_field_value;
use crate::vm::object::DvmObject;
use crate::vm::{DvmContext, DvmState};
use std::rc::Rc;

pub fn dm_print(args: Vec<DvmValue>, _state: &mut DvmState, _context: &DvmContext) -> DvmValue {
    if args.len() != 1 {
        return DvmValue::Void; // TODO: make exception
    }
    unsafe {
        print!("{}", get_string(&args[0]));
    }
    DvmValue::Void
}

pub fn dm_println(args: Vec<DvmValue>, _state: &mut DvmState, _context: &DvmContext) -> DvmValue {
    if args.len() != 1 {
        return DvmValue::Uninit;
    }
    unsafe {
        println!("{}", get_string(&args[0]));
    }
    DvmValue::Void
}

unsafe fn get_string(dvm_value: &DvmValue) -> String {
    match dvm_value {
        DvmValue::Byte(b) => b.to_string(),
        DvmValue::Int(i) => i.to_string(),
        DvmValue::Long(l) => l.to_string(),
        DvmValue::Double(d) => d.to_string(),
        DvmValue::Boolean(b) => b.to_string(),
        DvmValue::Object(object) => object_to_string(object.clone()),
        DvmValue::USize(u) => u.to_string(),
        DvmValue::Uninit => String::from("Uninit"),
        DvmValue::Void => String::from("Void"),
    }
}

fn object_to_string(alloc_object_rc: Rc<DvmObject>) -> String {
    if alloc_object_rc.implementation.fqn == "std::core::StringImpl" {
        extract_string_from_string(alloc_object_rc.clone())
    } else {
        todo!("what happens if we don't have a String?")
    }
}

fn extract_string_from_string(string_object: Rc<DvmObject>) -> String {
    let bytes_field = string_object
        .implementation
        .get_field("bytes", &string_object)
        .expect("Could not get String.bytes field.");
    let bytes_object = unsafe { get_field_value(&bytes_field, &string_object) }.expect_object();
    if bytes_object.implementation.fqn != "std::core::ArrayImpl" {
        panic!("String.bytes field is not a std::core::ArrayImpl");
    }

    let ptr_address_field = bytes_object
        .implementation
        .get_field("ptr_address", &bytes_object)
        .expect("Could not get ArrayImpl.ptr_address field.");
    let address = unsafe { get_field_value(&ptr_address_field, &bytes_object) }.expect_usize();

    let ptr_size_field = bytes_object
        .implementation
        .get_field("ptr_size", &bytes_object)
        .expect("Could not get ArrayImpl.ptr_size field.");
    let size = unsafe { get_field_value(&ptr_size_field, &bytes_object) }.expect_usize();

    let raw_bytes_pointer = address as *mut u8;
    let mut v: Vec<u8> = Vec::new();
    for i in 0..size {
        v.push(unsafe { raw_bytes_pointer.add(i).read() });
    }
    String::from_utf8(v).unwrap()
}
